@@ -5,19 +5,47 @@
 Utilities for sending files over ssh using the scp1 protocol.
 """
 
-__version__ = '0.13.0'
+__version__ = '0.14.4'
 
 import locale
 import os
 import re
 from socket import timeout as SocketTimeout
-import types
+
+
+SCP_COMMAND = b'scp'
+
+PATH_TYPES = (str, bytes)
+
+try:
+    import pathlib
+except ImportError:
+    pathlib = None
+else:
+    PATH_TYPES += pathlib.PurePath,
+try:
+    PATH_TYPES += unicode,
+except NameError:
+    pass
+
+try:
+    from typing import IO, TYPE_CHECKING, AnyStr, Callable, Iterable, Optional, Tuple, Union
+
+    if TYPE_CHECKING:
+        import paramiko.transport
+
+    # this is some magic to make sure pyright doesn't get too confused with pathlib potentially being a nullable variable
+    import pathlib
+    PurePath = pathlib.PurePath
+
+    PathTypes = Union[str, bytes, PurePath]
+except ImportError:
+    pass
 
 
 # this is quote from the shlex module, added in py3.3
 _find_unsafe = re.compile(br'[^\w@%+=:,./~-]').search
 
-
 def _sh_quote(s):
     """Return a shell-escaped version of the string `s`."""
     if not s:
@@ -39,6 +67,8 @@ def asbytes(s):
     """
     if isinstance(s, bytes):
         return s
+    elif pathlib and isinstance(s, pathlib.PurePath):
+        return bytes(s)
     else:
         return s.encode('utf-8')
 
@@ -86,6 +116,7 @@ class SCPClient(object):
     """
     def __init__(self, transport, buff_size=16384, socket_timeout=10.0,
                  progress=None, progress4=None, sanitize=_sh_quote):
+        # type: (paramiko.transport.Transport, int, float, Optional[Callable[[bytes, int, int], None]], Optional[Callable[[bytes, int, int, Tuple[str, int]], None]], Callable[[bytes], bytes]) -> None
         """
         Create an scp1 client.
 
@@ -117,11 +148,13 @@ class SCPClient(object):
         else:
             self._progress = None
         self._recv_dir = b''
+        self._depth = 0
         self._rename = False
         self._utime = None
         self.sanitize = sanitize
         self._dirtimes = {}
         self.peername = self.transport.getpeername()
+        self.scp_command = SCP_COMMAND
 
     def __enter__(self):
         self.channel = self._open()
@@ -132,6 +165,7 @@ class SCPClient(object):
 
     def put(self, files, remote_path=b'.',
             recursive=False, preserve_times=False):
+        # type: (Union[PathTypes, Iterable[PathTypes]], Union[str, bytes], bool, bool) -> None
         """
         Transfer files and directories to remote host.
 
@@ -151,13 +185,15 @@ class SCPClient(object):
         self.channel = self._open()
         self._pushed = 0
         self.channel.settimeout(self.socket_timeout)
-        scp_command = (b'scp -t ', b'scp -r -t ')[recursive]
+        scp_command = self.scp_command + b' ' + (b'-t ', b'-r -t ')[recursive]
         self.channel.exec_command(scp_command +
                                   self.sanitize(asbytes(remote_path)))
         self._recv_confirm()
 
-        if not isinstance(files, (list, tuple)):
+        if isinstance(files, PATH_TYPES):
             files = [files]
+        else:
+            files = list(files)
 
         if recursive:
             self._send_recursive(files)
@@ -167,6 +203,7 @@ class SCPClient(object):
         self.close()
 
     def putfo(self, fl, remote_path, mode='0644', size=None):
+        # type: (IO[AnyStr], Union[str, bytes], Union[str, bytes], Optional[int]) -> None
         """
         Transfer file-like object to remote host.
 
@@ -187,7 +224,7 @@ class SCPClient(object):
 
         self.channel = self._open()
         self.channel.settimeout(self.socket_timeout)
-        self.channel.exec_command(b'scp -t ' +
+        self.channel.exec_command(self.scp_command + b' -t ' +
                                   self.sanitize(asbytes(remote_path)))
         self._recv_confirm()
         self._send_file(fl, remote_path, mode, size=size)
@@ -195,6 +232,7 @@ class SCPClient(object):
 
     def get(self, remote_path, local_path='',
             recursive=False, preserve_times=False):
+        # type: (PathTypes, Union[str, bytes], bool, bool) -> None
         """
         Transfer files and directories from remote host to localhost.
 
@@ -210,10 +248,13 @@ class SCPClient(object):
             and directories.
         @type preserve_times: bool
         """
-        if not isinstance(remote_path, (list, tuple)):
+        if isinstance(remote_path, PATH_TYPES):
             remote_path = [remote_path]
+        else:
+            remote_path = list(remote_path)
         remote_path = [self.sanitize(asbytes(r)) for r in remote_path]
         self._recv_dir = local_path or os.getcwd()
+        self._depth = 0
         self._rename = (len(remote_path) == 1 and
                         not os.path.isdir(os.path.abspath(local_path)))
         if len(remote_path) > 1:
@@ -228,7 +269,7 @@ class SCPClient(object):
         self.channel = self._open()
         self._pushed = 0
         self.channel.settimeout(self.socket_timeout)
-        self.channel.exec_command(b"scp" +
+        self.channel.exec_command(self.scp_command +
                                   rcsv +
                                   prsv +
                                   b" -f " +
@@ -315,13 +356,15 @@ class SCPClient(object):
 
     def _send_recursive(self, files):
         for base in files:
+            base = asbytes(base)
             if not os.path.isdir(base):
                 # filename mixed into the bunch
                 self._send_files([base])
                 continue
             last_dir = asbytes(base)
             for root, dirs, fls in os.walk(base):
-                self._chdir(last_dir, asbytes(root))
+                if not asbytes(root).endswith(b'/'):
+                    self._chdir(last_dir, asbytes(root))
                 self._send_files([os.path.join(root, f) for f in fls])
                 last_dir = asbytes(root)
             # back out of the directory
@@ -377,7 +420,7 @@ class SCPClient(object):
             # wait for command as long as we're open
             self.channel.sendall('\x00')
             msg = self.channel.recv(1024)
-            if not msg:  # chan closed while recving
+            if not msg:  # chan closed while receiving
                 break
             assert msg[-1:] == b'\n'
             msg = msg[:-1]
@@ -410,11 +453,13 @@ class SCPClient(object):
                 path = self._recv_dir
                 self._rename = False
             elif os.name == 'nt':
-                path = os.path.join(asunicode_win(self._recv_dir),
-                                    parts[2].decode('utf-8'))
+                name = parts[2].decode('utf-8')
+                assert not os.path.isabs(name)
+                path = os.path.join(asunicode_win(self._recv_dir), name)
             else:
-                path = os.path.join(asbytes(self._recv_dir),
-                                    parts[2])
+                name = parts[2]
+                assert not os.path.isabs(name)
+                path = os.path.join(asbytes(self._recv_dir), name)
         except:
             chan.send('\x01')
             chan.close()
@@ -441,7 +486,10 @@ class SCPClient(object):
                 # we have to make sure we don't read the final byte
                 if size - pos <= buff_size:
                     buff_size = size - pos
-                file_hdl.write(chan.recv(buff_size))
+                data = chan.recv(buff_size)
+                if not data:
+                    raise SCPException("Underlying channel was closed")
+                file_hdl.write(data)
                 pos = file_hdl.tell()
                 if self._progress:
                     self._progress(path, size, pos, self.peername)
@@ -470,11 +518,15 @@ class SCPClient(object):
                 path = self._recv_dir
                 self._rename = False
             elif os.name == 'nt':
-                path = os.path.join(asunicode_win(self._recv_dir),
-                                    parts[2].decode('utf-8'))
+                name = parts[2].decode('utf-8')
+                assert not os.path.isabs(name)
+                path = os.path.join(asunicode_win(self._recv_dir), name)
+                self._depth += 1
             else:
-                path = os.path.join(asbytes(self._recv_dir),
-                                    parts[2])
+                name = parts[2]
+                assert not os.path.isabs(name)
+                path = os.path.join(asbytes(self._recv_dir), name)
+                self._depth += 1
         except:
             self.channel.send(b'\x01')
             raise SCPException('Bad directory format')
@@ -493,7 +545,9 @@ class SCPClient(object):
             raise
 
     def _recv_popd(self, *cmd):
-        self._recv_dir = os.path.split(self._recv_dir)[0]
+        if self._depth > 0:
+            self._depth -= 1
+            self._recv_dir = os.path.split(self._recv_dir)[0]
 
     def _set_dirtimes(self):
         try:
@@ -510,6 +564,7 @@ class SCPException(Exception):
 
 def put(transport, files, remote_path=b'.',
         recursive=False, preserve_times=False):
+    # type: (paramiko.transport.Transport, Union[PathTypes, Iterable[PathTypes]], Union[str, bytes], bool, bool) -> None
     """
     Transfer files and directories to remote host.
 
@@ -534,6 +589,7 @@ def put(transport, files, remote_path=b'.',
 
 def get(transport, remote_path, local_path='',
         recursive=False, preserve_times=False):
+    # type: (paramiko.transport.Transport, PathTypes, Union[str, bytes], bool, bool) -> None
     """
     Transfer files and directories from remote host to localhost.
 
